--
--  Copyright (C) 2022, AdaCore
--
--  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
--

--  Constants for named characters.

package VSS.Characters.Specials is

   pragma Preelaborate;

   Replacement_Character : constant Virtual_Character :=
     Virtual_Character'Val (16#FFFD#);

end VSS.Characters.Specials;
