--
--  Copyright (C) 2020, AdaCore
--
--  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
--

package VSS is

   pragma Pure;

   Magic_Error : exception;
   --  Raised in case of any unhandled error

end VSS;
