--
--  Copyright (C) 2021, AdaCore
--
--  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
--

with VSS.Strings.Cursors.Iterators.Lines;

package VSS.Strings.Line_Iterators renames VSS.Strings.Cursors.Iterators.Lines;
